package Font::TTF::Woff::PrivateData;

=head1 NAME

Font::TTF::Woff::PrivateData - WOFF Private data

=head1 DESCRIPTION

Currently a stub, thus read() results in read_dat()

=cut

use Font::TTF::Utils;
require Font::TTF::Table;

@ISA = qw(Font::TTF::Table);

1;

=head1 AUTHOR

Bob Hallissy. L<http://scripts.sil.org/FontUtils>.

=head1 LICENSING

Copyright (c) 1998-2016, SIL International (http://www.sil.org) 

This module is released under the terms of the Artistic License 2.0. 
For details, see the full text of the license in the file LICENSE.

=cut

